//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_wave.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_wave.hpp"

//############################################################################
LaserBoy_wave_header::LaserBoy_wave_header(const LaserBoy_wave_header& header)
                      : num_samples         (0                        ), // this is a copy ctor for the header of an empty file!
                        sample_rate         (header.sample_rate       ),
                        LaserBoy_wave_mode  (header.LaserBoy_wave_mode),
                        num_channels        (header.num_channels      ),
                        bits_per_sample     (header.bits_per_sample   ),
                        version             (LASERBOY_WAVE_VERSION    ),
                        parms               (                         )
{
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS)
        memcpy((void*)offset, (void*)header.offset, header.num_channels * sizeof(int));
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_SIGNAL_BIT_RESOLUTION)
        memcpy((void*)resolution, (void*)header.resolution, header.num_channels * sizeof(u_char));
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_SIGNAL_MATRIX)
    {
        memcpy((void*)signal_id, (void*)header.signal_id, header.num_channels * sizeof(short));
        memcpy((void*)LSB_tag  , (void*)header.LSB_tag  , header.num_channels * sizeof(short));
    }
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
    {
        parms.lit_dwell_overhang     = header.parms.lit_dwell_overhang    ;
        parms.lit_delta_max          = header.parms.lit_delta_max         ;
        parms.blank_delta_max        = header.parms.blank_delta_max       ;
        parms.max_dwell_microsec     = header.parms.max_dwell_microsec    ;
        parms.insignificant_distance = header.parms.insignificant_distance;
        parms.insignificant_angle    = header.parms.insignificant_angle   ;
        parms.frames_per_second      = header.parms.frames_per_second     ;
    }
}

//############################################################################
LaserBoy_wave_header::LaserBoy_wave_header(const LaserBoy_wave_header& header, int)
                      : num_samples         (0                        ), // this is a copy ctor for the header of an empty file!
                        sample_rate         (header.sample_rate       ),
                        LaserBoy_wave_mode  (header.LaserBoy_wave_mode),
                        num_channels        (8                        ),
                        bits_per_sample     (header.bits_per_sample   ),
                        version             (LASERBOY_WAVE_VERSION    ),
                        parms               (                         )
{
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS)
    {
        memcpy((void*)offset, (void*)header.offset, header.num_channels * sizeof(int));
        offset    [6] = 0;
        offset    [7] = 0;
    }
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_SIGNAL_BIT_RESOLUTION)
    {
        memcpy((void*)resolution, (void*)header.resolution, header.num_channels * sizeof(u_char));
        resolution[6] = 16;
        resolution[7] = 16;
    }
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_SIGNAL_MATRIX)
    {
        memcpy((void*)signal_id, (void*)header.signal_id, header.num_channels * sizeof(short));
        memcpy((void*)LSB_tag  , (void*)header.LSB_tag  , header.num_channels * sizeof(short));
        signal_id [6] = LASERBOY_SIGNAL_AUDIO_LEFT ;
        signal_id [7] = LASERBOY_SIGNAL_AUDIO_RIGHT;

        LSB_tag   [6] = LASERBOY_LSB_NOT_USED;
        LSB_tag   [7] = LASERBOY_LSB_NOT_USED;
    }
    //------------------------------------------------------------------------
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
    {
        parms.lit_dwell_overhang     = header.parms.lit_dwell_overhang;
        parms.lit_delta_max          = header.parms.lit_delta_max;
        parms.blank_delta_max        = header.parms.blank_delta_max;
        parms.max_dwell_microsec     = header.parms.max_dwell_microsec;
        parms.insignificant_distance = header.parms.insignificant_distance;
        parms.insignificant_angle    = header.parms.insignificant_angle;
        parms.frames_per_second      = header.parms.frames_per_second;
    }
}

//############################################################################
LaserBoy_wave_header::LaserBoy_wave_header(ifstream& in)
{
    bool    done = false    ;
    char    a, b, c, d      ;

    short   audio_format    ,
            block_align     ;

    int     i               ,
            chunk_size      ,
            sub_chunk_1_size,
            sub_chunk_2_size,
            byte_rate       ;

    string  temp_string     ,
            version_guess   ;
    //------------------------------------------------------------------------
    version            = "!LaserBoy!"; // assume the worst!
    LaserBoy_wave_mode = LASERBOY_WAVE_NO_MODE;
    //------------------------------------------------------------------------
    do
    {
        in.get(a); if(a != 'R') break;
        in.get(a); if(a != 'I') break;
        in.get(a); if(a != 'F') break;
        in.get(a); if(a != 'F') break;
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        chunk_size = (   (a & 0x000000ff)
                       | (b & 0x000000ff) << 8
                       | (c & 0x000000ff) << 16
                       | (d & 0x000000ff) << 24
                     );
        //--------------------------------------------------------------------
        in.get(a); if(a != 'W') break;
        in.get(a); if(a != 'A') break;
        in.get(a); if(a != 'V') break;
        in.get(a); if(a != 'E') break;
        //--------------------------------------------------------------------
        in.get(a); if(a != 'f') break;
        in.get(a); if(a != 'm') break;
        in.get(a); if(a != 't') break;
        in.get(a); if(a != ' ') break;
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sub_chunk_1_size = (   (a & 0x000000ff)
                             | (b & 0x000000ff) << 8
                             | (c & 0x000000ff) << 16
                             | (d & 0x000000ff) << 24
                           );
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        audio_format = ((b & 0x00ff) << 8 | (a & 0x00ff));
        if(audio_format != 1) break;
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        num_channels = ((b & 0x00ff) << 8 | (a & 0x00ff));
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sample_rate = (   (a & 0x000000ff)
                        | (b & 0x000000ff) << 8
                        | (c & 0x000000ff) << 16
                        | (d & 0x000000ff) << 24
                      );
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        byte_rate = (   (a & 0x000000ff)
                      | (b & 0x000000ff) << 8
                      | (c & 0x000000ff) << 16
                      | (d & 0x000000ff) << 24
                    );
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        block_align = ((b & 0x00ff) << 8 | (a & 0x00ff));
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        bits_per_sample = ((b & 0x00ff) << 8 | (a & 0x00ff));
        //--------------------------------------------------------------------
        if(byte_rate   != (sample_rate * num_channels * (bits_per_sample / 8)))
            break;
        if(block_align != (num_channels * (bits_per_sample / 8)))
            break;
        //--------------------------------------------------------------------
        if(sub_chunk_1_size > 16) // something has been added to this header
        {
            if(num_channels >= 6)
            {
                LaserBoy_wave_mode =   LASERBOY_WAVE_POSITIVE
                                     | LASERBOY_WAVE_END_OF_FRAME
                                     | LASERBOY_WAVE_SIGNAL_MATRIX;

                signal_id[0] = LASERBOY_SIGNAL_X_POSITION  ; // default values
                signal_id[1] = LASERBOY_SIGNAL_Y_POSITION  ;
                signal_id[2] = LASERBOY_SIGNAL_RED_ANALOG  ;
                signal_id[3] = LASERBOY_SIGNAL_GREEN_ANALOG;
                signal_id[4] = LASERBOY_SIGNAL_BLUE_ANALOG ;
                signal_id[5] = LASERBOY_SIGNAL_UNDEFINED   ;
                signal_id[6] = LASERBOY_SIGNAL_UNDEFINED   ;
                signal_id[7] = LASERBOY_SIGNAL_UNDEFINED   ;

                LSB_tag  [0] = LASERBOY_LSB_NOT_USED    ; // default values
                LSB_tag  [1] = LASERBOY_LSB_NOT_USED    ;
                LSB_tag  [2] = LASERBOY_LSB_END_OF_FRAME;
                LSB_tag  [3] = LASERBOY_LSB_NOT_USED    ;
                LSB_tag  [4] = LASERBOY_LSB_NOT_USED    ;
                LSB_tag  [5] = LASERBOY_LSB_NOT_USED    ;
                LSB_tag  [6] = LASERBOY_LSB_NOT_USED    ;
                LSB_tag  [7] = LASERBOY_LSB_NOT_USED    ;

                resolution[0] = 16; // default values
                resolution[1] = 16;
                resolution[2] =  8;
                resolution[3] =  8;
                resolution[4] =  8;
                resolution[5] = 16;
                resolution[6] = 16;
                resolution[7] = 16;

                if(num_channels == 8)
                {
                    signal_id[6] = LASERBOY_SIGNAL_AUDIO_LEFT ; // default values
                    signal_id[7] = LASERBOY_SIGNAL_AUDIO_RIGHT;
                }
            }
            //----------------------------------------------------------------
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            //----------------------------------------------------------------
            if(temp_string == "offsets ")
            {
                LaserBoy_wave_mode |= LASERBOY_WAVE_OFFSETS;
                //------------------------------------------------------------
                for(i = 0; i < num_channels; i++)
                {
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    offset[i] = (   (a & 0x000000ff)
                                  | (b & 0x000000ff) << 8
                                  | (c & 0x000000ff) << 16
                                  | (d & 0x000000ff) << 24
                                );
                }
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                if(    (a == 'd')
                    && (b == 'a')
                    && (c == 't')
                    && (d == 'a')
                  )
                {
                    version_guess = "v0 offsets       ";
                } // end if( we just read "data" )
                //------------------------------------------------------------
                else
                {
                    if(    (a == 'e') // we may or may not have an "eof3" mark
                        && (b == 'o') // if we do we don't have LaserBoy_wave_mode
                        && (c == 'f') // LaserBoy_wave_mode has default values set in the function
                        && (d == '3')
                      )
                    {
                        version_guess = "v1 offsets eof3  ";
                    } // end if( we just read "eof3" )
                    //--------------------------------------------------------
                    else
                    {
                        LaserBoy_wave_mode = (   (a & 0x000000ff)
                                               | (b & 0x000000ff) << 8
                                               | (c & 0x000000ff) << 16
                                               | (d & 0x000000ff) << 24
                                             );
                        version_guess = "v2 LaserBoy_wave_mode  ";

                    } // end else we take this as LaserBoy_wave_mode
                    //--------------------------------------------------------
                    LaserBoy_wave_mode |= LASERBOY_WAVE_OFFSETS; // we know we have offsets
                    //--------------------------------------------------------
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    //--------------------------------------------------------
                    if(a != 'd') break;
                    if(b != 'a') break;
                    if(c != 't') break;
                    if(d != 'a') break;
                }
            } // end if(temp_string == "offsets ")
            //----------------------------------------------------------------
            else if(temp_string == "offsetsZ")
            {
                LaserBoy_wave_mode |= LASERBOY_WAVE_OFFSETS;
                //------------------------------------------------------------
                for(i = 0; i < num_channels; i++)
                {
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    offset[i] = (   (a & 0x000000ff)
                                  | (b & 0x000000ff) << 8
                                  | (c & 0x000000ff) << 16
                                  | (d & 0x000000ff) << 24
                                );
                }
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                if(    (a == 'd')
                    && (b == 'a')
                    && (c == 't')
                    && (d == 'a')
                  )
                {
                    version_guess = "v0a offsetsZ     ";
                    signal_id [5] = LASERBOY_SIGNAL_Z_POSITION;
                    resolution[5] = 16;
                } // end if( we just read "data" )
                //------------------------------------------------------------
            } // end else if(temp_string == "offsetsZ")
            //----------------------------------------------------------------
            else if(temp_string == "LaserBoy")
            {
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                //------------------------------------------------------------
                if(    temp_string != "LaserBoy06072008"
                    && temp_string != "LaserBoy07132008"
                    && temp_string != "LaserBoy10102008"
                    && temp_string != LASERBOY_WAVE_VERSION
                  )
                    break;
                //------------------------------------------------------------
                version_guess = temp_string;
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                LaserBoy_wave_mode = (   (a & 0x000000ff)
                                       | (b & 0x000000ff) << 8
                                       | (c & 0x000000ff) << 16
                                       | (d & 0x000000ff) << 24
                                     );
                //------------------------------------------------------------
                if(LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS)
                    for(i = 0; i < num_channels; i++)
                    {
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        offset[i] = (   (a & 0x000000ff)
                                      | (b & 0x000000ff) << 8
                                      | (c & 0x000000ff) << 16
                                      | (d & 0x000000ff) << 24
                                    );
                    }
                //------------------------------------------------------------
                else
                {
                    LaserBoy_wave_mode |= LASERBOY_WAVE_OFFSETS;
                    for(i = 0; i < num_channels; i++)
                        offset[i] = 0;
                }
                //------------------------------------------------------------
                if(LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
                {
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    parms.lit_dwell_overhang = (   (a & 0x000000ff)
                                                 | (b & 0x000000ff) << 8
                                                 | (c & 0x000000ff) << 16
                                                 | (d & 0x000000ff) << 24
                                               );
                    //--------------------------------------------------------
                    if(version_guess == "LaserBoy06072008")
                    {   // floats were stored backwards. Oops!
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.lit_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        parms.blank_delta_max = parms.lit_delta_max;
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.max_dwell_microsec = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.insignificant_distance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.insignificant_angle = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.frames_per_second = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                    }
                    //--------------------------------------------------------
                    else if(version_guess == "LaserBoy10102008")
                    {
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.lit_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        parms.blank_delta_max = parms.lit_delta_max;
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.max_dwell_microsec = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.insignificant_distance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.insignificant_angle = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.frames_per_second = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                    }
                    //--------------------------------------------------------
                    else
                    {
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.lit_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.blank_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.max_dwell_microsec = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.insignificant_distance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.insignificant_angle = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.frames_per_second = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                    }
                }
                //------------------------------------------------------------
                if(LaserBoy_wave_mode & LASERBOY_WAVE_SIGNAL_MATRIX)
                    for(i = 0; i < num_channels; i++)
                    {
                        in.get(a); in.get(b);
                        signal_id[i] = ((b & 0x00ff) << 8 | (a & 0x00ff));
                        //----------------------------------------------------
                        in.get(a); in.get(b);
                        LSB_tag  [i] = ((b & 0x00ff) << 8 | (a & 0x00ff));
                        //----------------------------------------------------
                        if(signal_id[5] == LASERBOY_SIGNAL_Z_POSITION)
                            resolution[5] = 16;
                    }
                //------------------------------------------------------------
                else
                {
                    if(num_channels >= 6)
                    {
                        signal_id [5] = LASERBOY_SIGNAL_MONO_TTL ;
                        LSB_tag   [3] = LASERBOY_LSB_UNIQUE_FRAME;
                    }
                    else
                        for(i = 0; i < num_channels; i++)
                        {
                            signal_id[i] = LASERBOY_SIGNAL_UNDEFINED;
                            LSB_tag  [i] = LASERBOY_LSB_NOT_USED    ;
                        }
                }
                //------------------------------------------------------------
                if(LaserBoy_wave_mode & LASERBOY_SIGNAL_BIT_RESOLUTION)
                    for(i = 0; i < num_channels; i++)
                    {
                        in.get(a);
                        resolution[i] = (u_char)a;
                    }
                else
                {
                    if(num_channels >= 6)
                        resolution[5] = 8;
                }
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                if(a != 'd') break;
                if(b != 'a') break;
                if(c != 't') break;
                if(d != 'a') break;
            //----------------------------------------------------------------
            } // end else if(temp_string == "LaserBoy")
            else
                break;
        } // end if(sub_chunk_1_size > 16)
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sub_chunk_2_size = (   (a & 0x000000ff)
                             | (b & 0x000000ff) << 8
                             | (c & 0x000000ff) << 16
                             | (d & 0x000000ff) << 24
                           );
        //if(sub_chunk_2_size != (chunk_size + 36)) break;
        //--------------------------------------------------------------------
        num_samples = sub_chunk_2_size / (num_channels * (bits_per_sample / 8));
        version     = version_guess; // Woohoo !!!
        done        = true;
    } while(!done);
}

//############################################################################
void LaserBoy_wave_header::to_ofstream_wave(ofstream& out) const
{
    short   audio_format     = 1,
            block_align      = num_channels * (bits_per_sample / 8);

    int     i,
            sub_chunk_1_size = 16, // standard wave Subchunk1Size
            sub_chunk_2_size = num_samples * num_channels * (bits_per_sample / 8),
            byte_rate        = sample_rate * num_channels * (bits_per_sample / 8),
            chunk_size       = 20 + sub_chunk_1_size + sub_chunk_2_size;
    //------------------------------------------------------------------------
    if(LaserBoy_wave_mode != LASERBOY_WAVE_NO_MODE)
        sub_chunk_1_size += (   16 // sizeof "LaserBoymmddCCYY"
                              + sizeof(int) // sizeof (int)LaserBoy_wave_mode
                              + (   (LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS)
                                  ? (num_channels * sizeof(int))
                                  : (0)
                                )
                              + (   (LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
                                  ? (7 * sizeof(int)) // seven 32bit numbers
                                  : (0)
                                )
                              + (   (LaserBoy_wave_mode & LASERBOY_WAVE_SIGNAL_MATRIX)
                                  ? (num_channels * 2 * sizeof(short))
                                  : (0)
                                )
                              + (   (LaserBoy_wave_mode & LASERBOY_SIGNAL_BIT_RESOLUTION)
                                  ? (num_channels * sizeof(u_char))
                                  : (0)
                                )
                            );
    //------------------------------------------------------------------------
    out.seekp(0, ios::beg);
    //------------------------------------------------------------------------
    out.put('R');
    out.put('I');
    out.put('F');
    out.put('F');

    out.put((char) (chunk_size & 0x000000ff)       );
    out.put((char)((chunk_size & 0x0000ff00) >> 8 ));
    out.put((char)((chunk_size & 0x00ff0000) >> 16));
    out.put((char)((chunk_size & 0xff000000) >> 24));

    out.put('W');
    out.put('A');
    out.put('V');
    out.put('E');

    out.put('f');
    out.put('m');
    out.put('t');
    out.put(' ');

    out.put((char) (sub_chunk_1_size & 0x000000ff)       );
    out.put((char)((sub_chunk_1_size & 0x0000ff00) >> 8 ));
    out.put((char)((sub_chunk_1_size & 0x00ff0000) >> 16));
    out.put((char)((sub_chunk_1_size & 0xff000000) >> 24));

    out.put((char) (audio_format & 0x00ff)      );
    out.put((char)((audio_format & 0xff00) >> 8)); // 2

    out.put((char) (num_channels & 0x00ff)      );
    out.put((char)((num_channels & 0xff00) >> 8)); // 4

    out.put((char) (sample_rate & 0x000000ff)       );
    out.put((char)((sample_rate & 0x0000ff00) >> 8 ));
    out.put((char)((sample_rate & 0x00ff0000) >> 16));
    out.put((char)((sample_rate & 0xff000000) >> 24)); // 8

    out.put((char) (byte_rate & 0x000000ff)       );
    out.put((char)((byte_rate & 0x0000ff00) >> 8 ));
    out.put((char)((byte_rate & 0x00ff0000) >> 16));
    out.put((char)((byte_rate & 0xff000000) >> 24)); // 12

    out.put((char) (block_align & 0x00ff)      );
    out.put((char)((block_align & 0xff00) >> 8)); // 14

    out.put((char) (bits_per_sample & 0x00ff)      );
    out.put((char)((bits_per_sample & 0xff00) >> 8)); // 16
//----------------------------------------------------------------------------
// extra structures added for LaserBoy !!!
//----------------------------------------------------------------------------
    if(LaserBoy_wave_mode != LASERBOY_WAVE_NO_MODE)
    {
        out.put(version[ 0]); // L
        out.put(version[ 1]); // a
        out.put(version[ 2]); // s
        out.put(version[ 3]); // e
        out.put(version[ 4]); // r
        out.put(version[ 5]); // B
        out.put(version[ 6]); // o
        out.put(version[ 7]); // y

        out.put(version[ 8]); // M
        out.put(version[ 9]); // M
        out.put(version[10]); // D
        out.put(version[11]); // D
        out.put(version[12]); // C
        out.put(version[13]); // C
        out.put(version[14]); // Y
        out.put(version[15]); // Y
        //--------------------------------------------------------------------
        out.put((char) (LaserBoy_wave_mode & 0x000000ff)       );
        out.put((char)((LaserBoy_wave_mode & 0x0000ff00) >> 8 ));
        out.put((char)((LaserBoy_wave_mode & 0x00ff0000) >> 16));
        out.put((char)((LaserBoy_wave_mode & 0xff000000) >> 24)); // 36
        //--------------------------------------------------------------------
        if(LaserBoy_wave_mode & LASERBOY_WAVE_OFFSETS)
            for(i = 0; i < num_channels; i++)
            {
                out.put((char)( offset[i] & 0x000000ff)       );
                out.put((char)((offset[i] & 0x0000ff00) >> 8 ));
                out.put((char)((offset[i] & 0x00ff0000) >> 16));
                out.put((char)((offset[i] & 0xff000000) >> 24)); // 36 + 4 * num_channels
            }
        //--------------------------------------------------------------------
        if(LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
        {
            out.put((char) (parms.lit_dwell_overhang & 0x000000ff)       ); // int
            out.put((char)((parms.lit_dwell_overhang & 0x0000ff00) >> 8 ));
            out.put((char)((parms.lit_dwell_overhang & 0x00ff0000) >> 16));
            out.put((char)((parms.lit_dwell_overhang & 0xff000000) >> 24));

            out.put((char)(((int)(parms.lit_delta_max) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.lit_delta_max) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.lit_delta_max) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.lit_delta_max) & 0x000000ff)       );

            out.put((char)(((int)(parms.blank_delta_max) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.blank_delta_max) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.blank_delta_max) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.blank_delta_max) & 0x000000ff)       );

            out.put((char)(((int)(parms.max_dwell_microsec) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.max_dwell_microsec) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.max_dwell_microsec) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.max_dwell_microsec) & 0x000000ff)       );

            out.put((char)(((int)(parms.insignificant_distance) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.insignificant_distance) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.insignificant_distance) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.insignificant_distance) & 0x000000ff)       );

            out.put((char)(((int)(parms.insignificant_angle) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.insignificant_angle) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.insignificant_angle) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.insignificant_angle) & 0x000000ff)       );

            out.put((char)(((int)(parms.frames_per_second) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.frames_per_second) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.frames_per_second) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.frames_per_second) & 0x000000ff)       );
        }
        //--------------------------------------------------------------------
        if(LaserBoy_wave_mode & LASERBOY_WAVE_SIGNAL_MATRIX)
            for(i = 0; i < num_channels; i++)
            {
                out.put((char) (signal_id[i] & 0x00ff)      );
                out.put((char)((signal_id[i] & 0xff00) >> 8));

                out.put((char) (LSB_tag[i] & 0x00ff)      );
                out.put((char)((LSB_tag[i] & 0xff00) >> 8));
            }
        //--------------------------------------------------------------------
        if(LaserBoy_wave_mode & LASERBOY_SIGNAL_BIT_RESOLUTION)
            for(i = 0; i < num_channels; i++)
                out.put(resolution[i]);
    } // end if(with_LaserBoy_stuff)
    //------------------------------------------------------------------------
    out.put('d');
    out.put('a');
    out.put('t');
    out.put('a');
    //------------------------------------------------------------------------
    out.put((char) (sub_chunk_2_size & 0x000000ff)       );
    out.put((char)((sub_chunk_2_size & 0x0000ff00) >> 8 ));
    out.put((char)((sub_chunk_2_size & 0x00ff0000) >> 16));
    out.put((char)((sub_chunk_2_size & 0xff000000) >> 24));

    return;
}


//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
